/*!
 * @file	keypaddev.c
 * FILE: $Id: keypaddev.c,v 1.1.1.1 2007/03/06 04:50:59 smcdaid Exp $
 *
 * DESCRIPTION:
 *     Simple button and rotatory encoder device driver which interfaces 
 *     a keypad. The keypad consists of LEFT, RIGHT, UP, DOWN, CENTER 
 *     buttons and a rotatory encoder. It is interfaced through GPIO on
 *     Xilinx Spartan3 PetaLogix FPGA board.
 *
 * AUTHOR: 
 *     Irfan Syed <irfan@itee.uq.edu.au>
 *
 * LICENSING:
 *     Copyright (c) 2006 PetaLogix. All rights reserved.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */


/* -- required headers -- */
#include <linux/module.h>
#include <linux/init.h>

#include <asm/irq.h>
#include <asm/io.h>

#include <linux/interrupt.h>
#include <linux/types.h> /* u32 */

#include <linux/autoconf.h> /* hardware definitions */

#include <linux/input.h>

/* #define DEBUG */

/* -- module requirements -- */
MODULE_LICENSE( "GPL" );
MODULE_DESCRIPTION( "Keypaddev Device Driver" );
//MODULE_VERSION( "1.0" );
MODULE_AUTHOR( "Irfan Syed" );


/* -- function prototypes -- */
static irqreturn_t btn_interrupt( int , void *, struct pt_regs * );
static int __init btn_interf_init( void );
static void __exit btn_interf_exit( void );

/* Register init, exit functions */
module_init( btn_interf_init );
module_exit( btn_interf_exit );


/* -- global private variables -- */
#define INPUT_OFFSET	0x00000000
#define KPENTER_MASK	0x00000001
#define UP_MASK			0x00000002
#define DOWN_MASK		0x00000004
#define LEFT_MASK		0x00000008
#define RIGHT_MASK		0x00000010
#define ROT_RIGHT_MASK	0x00000020
#define ROT_LEFT_MASK	0x00000040

/* hardware address definitions */
#define C_BASEADDR		(CONFIG_XILINX_BTN_DECODER_0_BASEADDR)
#define GPIO_DATA		(C_BASEADDR + INPUT_OFFSET)
#define OPB_GIE			(C_BASEADDR + 0x11c)
#define IP_IER			(C_BASEADDR + 0x128)
#define IP_ISR			(C_BASEADDR + 0x120)

#define C_INTERRUPT_PRESENT	1 /* to enable interrupts */

/* enabling masks */
#define GIE_MASK		(1<<31)		/* Global interrupt enable bit */
#define IP_IER_MASK		(1<<0)		/* Channel 1 enable */
#define IP_ISR_MASK		(1<<0)		/* Channel 1 ack */

volatile u32 *gpio_data_ptr = (u32 *)GPIO_DATA;
volatile u32 *opb_gie_ptr = (u32 *)OPB_GIE;
volatile u32 *ip_ier_ptr = (u32 *)IP_IER;
volatile u32 *ip_isr_ptr = (u32 *)IP_ISR;

/* button device */
static struct input_dev btn_dev = {
		.evbit	= { BIT(EV_KEY) | BIT(EV_REL) },
		.keybit	= { [LONG(KEY_LEFT)] = BIT(KEY_LEFT) | BIT(KEY_RIGHT) | 
				BIT(KEY_UP) | BIT(KEY_DOWN) | BIT(KEY_KPENTER) },
		.relbit = { BIT(REL_DIAL) },
		.name	= "Button Keypad",
		.idbus = BUS_HOST,
		.idvendor = 0x0001,
		.idproduct = 0x0001,
		.idversion = 0x1111,
};

static u32 raw_prev = 0x00000000;
static u32 raw_new = 0x00000000;

static u32 old_rot_left = 0x00000000;
static u32 old_rot_right = 0x00000000;
static u32 new_rot_left = 0x00000000;
static u32 new_rot_right = 0x00000000;


/* -- function implementations -- */
/*! Interrupt Handler: btn_interrupt
 * 		Generates interrupts (LEFT, RIGHT, UP, DOWN, KPENTER to the 
 * 		event subsystem.
 *
 * 	@param	irq:
 * 	@param	dev_id:
 * 	@param	fp:
 *
 * 	@return	IRQ_HANDLED
 */
static irqreturn_t btn_interrupt( int irq, void *dev_id, struct pt_regs *fp )
{
	/* Make sure this is our interrupt */
	if( !ip_isr_ptr[0] & IP_ISR_MASK )
		return IRQ_NONE;

	/* Read the raw GPIO value */
	raw_new = gpio_data_ptr[0];

#ifdef DEBUG
	printk( KERN_ALERT "value read from location 0x%08X=0x%08X\n", (GPIO_DATA), 
			raw_new );
#endif
	if( raw_prev != raw_new ) {
#ifdef DEBUG
	printk( KERN_ALERT "new_value=0x%08X, old_value=0x08%X\n", raw_new, 
			raw_prev );
#endif
		raw_prev = raw_new;
		/* process read input */
		/* buttons */
		input_report_key( &btn_dev, KEY_LEFT, (raw_prev & LEFT_MASK) );
		input_report_key( &btn_dev, KEY_RIGHT, (raw_prev & RIGHT_MASK) );
		input_report_key( &btn_dev, KEY_UP, (raw_prev & UP_MASK) );
		input_report_key( &btn_dev, KEY_DOWN, (raw_prev & DOWN_MASK) );
		input_report_key( &btn_dev, KEY_KPENTER, (raw_prev & KPENTER_MASK) );

		/* dial input */
		if( (new_rot_left = (raw_prev & ROT_LEFT_MASK)) != old_rot_left ) {
#ifdef DEBUG
	printk( "CCW\n" );
#endif
			old_rot_left = new_rot_left;
			input_report_rel( &btn_dev, REL_DIAL, -1 );
		}

		if( (new_rot_right = (raw_prev & ROT_RIGHT_MASK)) != old_rot_right ) {
#ifdef DEBUG
	printk( "CW\n" );
#endif
			old_rot_right = new_rot_right;
			input_report_rel( &btn_dev, REL_DIAL, 1 );
		}
	}

	/* Hit the ISR bit to clear interrupt status */
	ip_isr_ptr[0] |= IP_ISR_MASK;

	return IRQ_HANDLED;
}

/*! Module Initialisation: btn_interf_init
 * 		Implements initialisation of the the module when loaded.
 *
 * 	@param	NONE
 * 	
 * 	@return	0 if successful, -ive error code otherwise
 */
static int __init btn_interf_init( void )
{
	if( request_irq(CONFIG_XILINX_BTN_DECODER_0_IRQ, btn_interrupt, 0, "PetaLogix keypad device", 
			NULL) )	{
		printk( KERN_ERR "keypaddev.c: can't allocate irq %d\n", 
				CONFIG_XILINX_BTN_DECODER_0_IRQ );
		return -EBUSY;
	}

	input_register_device( &btn_dev );

	/* Set global interrupt enable on the device*/
	opb_gie_ptr[0] |= GIE_MASK;
	/* And interrupt enable on channel 1 */
	ip_ier_ptr[0] |= IP_IER_MASK;

#ifdef DEBUG
	printk( KERN_ALERT "HW interrupts enabled\n" );
#endif

	printk( KERN_INFO "Directional Button & Rotary Encoder Driver\n" );
	printk( KERN_INFO "(c) 2006 PetaLogix\n" );

	return 0; /* success */
}

/*! Module Exit: btn_interf_exit
 * 		Implements de-initialisation of the the module when un-loaded.
 *
 * 	@param	NONE
 *
 * 	@return	NONE
 */
static void __exit btn_interf_exit( void )
{
	/* disable hw interrupts */
	ip_ier_ptr[0] &= ~IP_IER_MASK;
	opb_gie_ptr[0] &= ~GIE_MASK;
#ifdef DEBUG
	printk( KERN_ALERT "HW interrupts dis-enabled\n" );
#endif

	free_irq(CONFIG_XILINX_BTN_DECODER_0_IRQ, btn_interrupt );
	input_unregister_device( &btn_dev );

	printk( KERN_INFO "Button Keypad un-initialised\n" );
}

