/**
 * \file md2.h
 *
 *  Based on XySSL: Copyright (C) 2006-2008  Christophe Devine
 *
 *  Copyright (C) 2009  Paul Bakker <polarssl_maintainer at polarssl dot org>
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of PolarSSL or XySSL nor the names of its contributors
 *      may be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef TROPICSSL_MD2_H
#define TROPICSSL_MD2_H

/**
 * \brief          MD2 context structure
 */
typedef struct {
	unsigned char cksum[16];	/*!< checksum of the data block */
	unsigned char state[48];	/*!< intermediate digest state  */
	unsigned char buffer[16];	/*!< data block being processed */

	unsigned char ipad[64];	/*!< HMAC: inner padding        */
	unsigned char opad[64];	/*!< HMAC: outer padding        */
	int left;		/*!< amount of data in buffer   */
} md2_context;

#ifdef __cplusplus
extern "C" {
#endif

	/**
	 * \brief          MD2 context setup
	 *
	 * \param ctx      context to be initialized
	 */
	void md2_starts(md2_context * ctx);

	/**
	 * \brief          MD2 process buffer
	 *
	 * \param ctx      MD2 context
	 * \param input    buffer holding the  data
	 * \param ilen     length of the input data
	 */
	void md2_update(md2_context * ctx, unsigned char *input, int ilen);

	/**
	 * \brief          MD2 final digest
	 *
	 * \param ctx      MD2 context
	 * \param output   MD2 checksum result
	 */
	void md2_finish(md2_context * ctx, unsigned char output[16]);

	/**
	 * \brief          Output = MD2( input buffer )
	 *
	 * \param input    buffer holding the  data
	 * \param ilen     length of the input data
	 * \param output   MD2 checksum result
	 */
	void md2(unsigned char *input, int ilen, unsigned char output[16]);

	/**
	 * \brief          Output = MD2( file contents )
	 *
	 * \param path     input file name
	 * \param output   MD2 checksum result
	 *
	 * \return         0 if successful, 1 if fopen failed,
	 *                 or 2 if fread failed
	 */
	int md2_file(char *path, unsigned char output[16]);

	/**
	 * \brief          MD2 HMAC context setup
	 *
	 * \param ctx      HMAC context to be initialized
	 * \param key      HMAC secret key
	 * \param keylen   length of the HMAC key
	 */
	void md2_hmac_starts(md2_context * ctx, unsigned char *key, int keylen);

	/**
	 * \brief          MD2 HMAC process buffer
	 *
	 * \param ctx      HMAC context
	 * \param input    buffer holding the  data
	 * \param ilen     length of the input data
	 */
	void md2_hmac_update(md2_context * ctx, unsigned char *input, int ilen);

	/**
	 * \brief          MD2 HMAC final digest
	 *
	 * \param ctx      HMAC context
	 * \param output   MD2 HMAC checksum result
	 */
	void md2_hmac_finish(md2_context * ctx, unsigned char output[16]);

	/**
	 * \brief          Output = HMAC-MD2( hmac key, input buffer )
	 *
	 * \param key      HMAC secret key
	 * \param keylen   length of the HMAC key
	 * \param input    buffer holding the  data
	 * \param ilen     length of the input data
	 * \param output   HMAC-MD2 result
	 */
	void md2_hmac(unsigned char *key, int keylen,
		      unsigned char *input, int ilen, unsigned char output[16]);

	/**
	 * \brief          Checkup routine
	 *
	 * \return         0 if successful, or 1 if the test failed
	 */
	int md2_self_test(int verbose);

#ifdef __cplusplus
}
#endif
#endif				/* md2.h */
