/*###########################################################################
#                                                                           #
#                          xnprintf() test suite                            #
#                                                                           #
#                  public domain      David TAILLANDIER                     #
#                 (only this file)                                          #
###########################################################################*/


/*
This test suite don't chech vnprintf() because there is not portable way
to compare two strings printed to screen.
This test suite don't check "non-v-versions" of functions (snprintf,
asprintf, ...) ; only "v-versions" are tested (vsnprintf, vasprintf, ...).
It is assumed that errors into non-tested functions are the same as in
other ones.
   tested functions:  vasprintf()
                      vasnprintf()
                      vsnprintf()
                      vfnprintf()

   the reference is:  vsprintf()  from _your_ standard library

exit value: 0 = ok
            1 = error
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "xnprintf.h"
#include "xnprintf.c"    /* just to avoid to have to link */

#define FILENAME "test.txt"

/* 16bit compilers and libraries under ms-dos can not handle big values */
#if defined _16BIT_ || defined __TINY__ || defined __SMALL__ || defined __MEDIUM__ || defined __COMPACT__ || defined __LARGE__ || defined __HUGE__
#    define __DOS_16BIT__
#endif

#if defined __TINY__ && defined __TURBOC__
#    error the math/float library do not work with TINY model
#endif


static unsigned int counter = 0;

/*################################ error ##################################*/
void
error( const char * text,
       ...
     )
{
     va_list vargs;

     fprintf( stderr, "\nERROR in xnprintf() test suite:\n" );
     va_start( vargs, text );
     vfprintf( stderr, text, vargs );
     va_end( vargs );

     fprintf( stderr, "\n" );

     exit( 1 );      /* 1 = error */
}

/*############################### compare #################################*/
void
compare( const char * info,
         int          maxlen,
         const char * format_string,
         ...
       )
{
     va_list vargs;
     char *  model;
     char *  test;
     int     retval;
     size_t  len;
     FILE *  file;

     counter++;
     
     if ( maxlen < 0 )
          error( "%s\n'maxlen' can not be a negative value", info );
     if ( maxlen == 0 )
          error( "%s\n'maxlen' can not be egal to 'zero' for tests", info );
     if ( maxlen == 1 )
          error( "%s\n'maxlen' can not be egal to 'one' for tests (for simplification)", info );

     va_start( vargs, format_string );

     /* uncomment this if you want to print to screen */
     /* vnprintf(maxlen,format_string,vargs);         */
     /* printf("\n");                                 */

     /* vasprintf */
     vasprintf( &model, format_string, vargs );
     if ( model == NULL )
          error( "%s\nNULL returned by vasprintf()", info );
     len = strlen( model );
     if ( strstr(model,"(null)") != NULL )
          error( "%s\nYou can not use NULL pointer for tests", info );
     /* because your library may handle NULL differently, */
     /* then nothing is possible to compare the results   */

     /* test string */
     test = (char *)malloc( len+1024 );    /* +1024 for safety, because this is a test suite, the goal is to find bugs */
     test[len+1] = 1;     test[len+2] = 2;     /* overflow marquers */
     test[len+3] = 3;     test[len+4] = 4;
     retval = vsprintf( test, format_string, vargs );
     if ( retval == EOF )
          error( "%s\nEOF returned by vsprintf()", info );
     if ( test[len+1]!=1 || test[len+2]!=2 || test[len+3]!=3 || test[len+4]!=4 )
          error( "%s\nduh ! overflow problem with native vsprintf()", info );

     /* check if 'model' and 'test' are identical */
     if ( strlen(test) != len )
          error( "%s\nvasprintf() gives a different length than vsprintf()", info );
     if ( strcmp(test,model) != 0 )
          error( "%s\nvasprintf() gives a different text than vsprintf()", info );

     /* cut the model */
     if ( (size_t)maxlen > len )
          maxlen = len + 1;
     model[maxlen-1] = 0;
     test[maxlen+0] = 1;     test[maxlen+1] = 2;     /* overflow marquers */
     test[maxlen+2] = 3;     test[maxlen+3] = 4;

     /* vsnprintf */
     if ( vsnprintf(test,maxlen,format_string,vargs) != retval )
          error( "%s\nvsnprintf() returns a different value than vsprinf()", info );
     if ( test[maxlen]!=1 || test[maxlen+1]!=2 || test[maxlen+2]!=3 || test[maxlen+3]!=4 )
          error( "%s\noverflow problem with vsnprintf()", info );
     if ( strcmp(test,model) != 0 )
          error( "%s\nvsnprintf() gives an erroneous text", info );

     /* vasnprintf */
       {
          char * buffer;

          if ( vasnprintf(&buffer,maxlen,format_string,vargs) != retval )
               error( "%s\nvasnprintf() returns a different value than vsprinf()", info );
          if ( strcmp(test,model) != 0 )
               error( "%s\nvasnprintf() gives an erroneous text", info );
          free( buffer );
       }

     /* vfnprintf */
     if ( fopen(FILENAME,"r") != NULL )
          error( "%s\nthe file %s already exists", info, FILENAME );

     file = fopen( FILENAME, "wb" );
     if ( file == NULL )
          error( "%s\nthe file %s can not be created", info, FILENAME );
     if ( vfnprintf(file,maxlen,format_string,vargs) != retval )
          error( "%s\nvfnprintf() returns a different value than vsprinf()", info );
     fclose( file );

     file = fopen( FILENAME, "rb" );
     if ( file == NULL )
          error( "%s\nthe file %s can not be opened", info, FILENAME );
     if ( fseek(file,0,SEEK_END) != 0 )
          error( "%s\nfseek() failed on %s", info, FILENAME );
     if ( ftell(file) != (long)(maxlen-1) )
          error( "%s\nthe %s file have a wrong size", info, FILENAME );
     if ( fseek(file,0,SEEK_SET) != 0 )
          error( "%s\nfseek() failed on %s", info, FILENAME );
     fread( test, 1, maxlen+100, file );     /* +100 to force to overflow */
     /* note: some library always give 'maxlen+100' as return */
     /* value for fread then we do not check this value       */
     fclose( file );
     if ( remove(FILENAME) != 0 )
          error( "%s\nthe file %s can not be erased", info, FILENAME );

     if ( test[maxlen]!=1 || test[maxlen+1]!=2 || test[maxlen+2]!=3 || test[maxlen+3]!=4 )
          error( "%s\noverflow problem with vfnprintf()", info );

     test[len] = 0;
     if ( strcmp(test,model) != 0 )
          error( "%s\nvfnprintf() gives an erroneous text", info );

     free( model );
     free( test );
     va_end( vargs );
}

/*################################# main ##################################*/
int
main( void )
{
     int i;
     int n1, n2;
     char dummy_str[50];

#define FORI(start,stop)  for(i=start;i!=stop;i++)

     FORI(2,30)     compare( "integer", i, "my_integer_is_%i", 1234 );
     FORI(2,30)     compare( "integer", i, "%i_is_my_integer", 1234 );
     FORI(2,30)     compare( "integer", i, "my_integer_is_%i,_right_?", 1234 );

#ifdef __DOS_16BIT__
     compare( "integer with HUGE format (huge)", 30000, "my_int_is_%6000i,_right_?", -1234 );
     FORI(6005,6015)   compare( "integer with HUGE format (huge,cut)", i, "my_int_is_%06000i,_right_?", -1234 );
#else
     compare( "integer with HUGE format (huge)", 30000, "my_int_is_%16000i,_right_?", -1234 );
     FORI(16005,16015)   compare( "integer with HUGE format (huge,cut)", i, "my_int_is_%016000i,_right_?", -1234 );
#endif
     FORI(2,30)      compare( "integer with special format", i, "my_int_%06i", 1234 );
     FORI(2,30)      compare( "integer with special format", i, "my_int_%06i", -1234 );
     FORI(2,30)      compare( "integer with special format", i, "my_int_%6.2i", 1234 );
     FORI(2,30)      compare( "integer with special format", i, "my_int_%6.2i", -1234 );

     FORI(2,30)      compare( "float", i, "my_float_%f", 123456.789 );
     FORI(2,30)      compare( "float", i, "my_float_is_%f,_right_?", 123456.789 );
     FORI(2,80)      compare( "float with special format", 300, "my_float_%025.10f", 123456.789 );
     FORI(2,80)      compare( "float with special format", 300, "my_float_%#05.10f", 123456.789 );
     FORI(2,80)      compare( "float with special format", 300, "my_float_is_%025.10f,_right_?", 123456.789 );
     FORI(2,80)      compare( "float with special format", 300, "my_float_is_%025.10f,_right_?", 123456.789 );
     FORI(2,80)      compare( "float with special format", i, "my_float_%25.4f", 123456.789 );
     FORI(2,80)      compare( "float with special format", i, "my_float_is_%5.40f,_right_?", 123456.789 );
     compare( "float (huge)", 30000, "my_float_%f", 123456.789 );
     compare( "float (huge)", 30000, "my_float_is_%f,_right_?", 123456.789 );
#ifdef __DOS_16BIT__
     compare( "huge float with special format", 30000, "my_float_is_%06000.1000f,_right_?", 123456.789 );
     FORI(6000,6050)    compare( "huge float with special format", i, "my_float_is_%06000.1000f,_right_?", 123456.789 );
#else
     compare( "huge float with special format (full)", 30000, "my_float_is_%016000.10000f,_right_?", 123456.789 );
     FORI(16000,16050)    compare( "huge float with special forma", i, "my_float_is_%016000.10000f,_right_?", 123456.789 );
#endif

     compare( "empty format", 10, "" );
     compare( "simple text", 10, "1234" );
     FORI(2,10)      compare( "simple text", i, "1234" );
     compare( "simple text (huge)", 30000, "1234" );

     compare( "empty string", 10, "%s", "" );
     compare( "empty string (huge)", 30000, "%s", "" );
     compare( "string (huge)", 30000, "%s", "1234" );
     FORI(2,11)      compare( "string", i, "%s", "1234567" );
#ifdef __DOS_16BIT__
     FORI(5990,6010)      compare( "huge string", i, "%06000s", "1234567" );
     FORI(5990,6010)      compare( "huge string", i, "%06000.4s", "1234567" );
#else
     FORI(15990,16010)      compare( "huge string", i, "%016000s", "1234567" );
     FORI(15990,16010)      compare( "huge string", i, "%016000.4s", "1234567" );
#endif

     compare( "writing to pointer", 30, "1234%n5678", &n1 );
     sprintf( dummy_str, "1234%n5678", &n2 );
     if ( n1 != n2 )
          error( "writing to pointer\nsprintf() returns a different int" );

     compare( "pointer (full)", 30, "abcd_%p", &i );
     FORI(4,20)   compare( "pointer", i, "abcd_%p", &i );

     FORI(2,100)
       {
          compare( "big one", i, "blah_%i%s%%_%5.9e%n1234%n___%10.1s\t_", 123, "abc", 123456.789, &n1, &n2, "abc" );
          if ( n1+4 != n2 )
               error( "big one\nproblem with n1 and n2" );
       }

     printf( "\nxnprintf() test suite: everything seems ok\n" );
     printf( "%u tests done.\n", counter );
     return 0;    /* 0 = ok */
}

/*############################# END OF FILE ###############################*/
